'use strict';

const NEW_TAB_URL = "chrome://newtab/";
const EXTENSIONS = "chrome://extensions/";
const PAGES = {
  APP: chrome.extension.getURL("pages/app.html"),
  OPTIONS: chrome.extension.getURL("pages/options.html"),
  DASHBOARD: chrome.extension.getURL("pages/dashboard.html"),
  ONBOARDING: chrome.extension.getURL("pages/onboarding.html")
}
// const API_HOST = "https://app.thyself.me";
const API_HOST = "https://thyself-adonis.herokuapp.com"
// const API_HOST = "http://127.0.0.1:3333";

chrome.runtime.onInstalled.addListener(function(details) {
  if( details.reason == "install" ){
    setNextPromptTime( currentTime() );
    setDisplayStrategy('activeTab');
    setDisplayDashboardOnNewTab(true);
    openInstallationPage();
  }
})

chrome.tabs.onCreated.addListener(function( tab ) {
  setLastHost(tab.id, tab.url);
})

chrome.tabs.onRemoved.addListener(function( tabId ) {
  deleteLastHost(tabId);
})

/*
 * main listeners – listen to tab changes to work out if Thyself should display or not
 */
chrome.tabs.onUpdated.addListener( async function( tabId, changeInfo, tab ) {

  if( changeInfo &&
      changeInfo.status === 'loading' &&
      !tab.url.includes( EXTENSIONS ) &&
      !tab.url.includes( PAGES.APP ) &&
      !tab.url.includes( PAGES.OPTIONS ) &&
      !tab.url.includes( PAGES.ONBOARDING ) &&
      !tab.url.includes( "haveagreatday" ) ) {

    const withinTabAllowance = await checkWithinTabAllowance()
    const hasPastPromptTime = await checkPastPromptTime()
    const hasChangedWebsite = await checkWebsiteChange( tab.id, tab.url )
    setLastHost(tab.id, tab.url)

    if( withinTabAllowance ) {
      let showingMoodSelection = false

      if( hasPastPromptTime ) {
        showingMoodSelection = await showMoodSelection( tab, hasChangedWebsite )
      }
      
      if( !showingMoodSelection ) {
        await showNewTabDashboard( tab )
      }
    }
  }
});

/*
 * showMoodSelection – try show the mood selection screens
 */
async function showMoodSelection( tab, hasChangedWebsite ) {
  let pageUrl = `${PAGES.APP}?url=${tab.url}`;
  const displayStrategy = await getDisplayStrategy();

  if( displayStrategy === 'newTab' ) {
    if( isOnNewTab( tab ) ) {
      if( await checkConnection() ) {
        showApp( tab, pageUrl )
        return true
      }
    }
  } else if( displayStrategy === 'activeTab' ) {
    if( await isOnChangedActiveTab( tab, hasChangedWebsite ) ) {
      if( await checkConnection() ) {
        showApp( tab, pageUrl )
        return true
      }
    }
  }

  return false
}

/*
 * showMoodSelection – try show the Thyself dashboard on the new tab screen 
 */
async function showNewTabDashboard( tab ) {
  if( await shouldShowDashboardOnNewTab( tab ) ) {
    if( await checkConnection() ) {
      const pageUrl = `${PAGES.DASHBOARD}?haveagreatday`
      showApp( tab, pageUrl )
    }
  }
}

/*
 * isOnNewTab – check if the page is the new tab
 */
function isOnNewTab( tab ) {
  return tab.url.includes( NEW_TAB_URL )
}

/*
 * isOnChangedActiveTab – check if we're on an active tab and it has changed website
 */
async function isOnChangedActiveTab( tab, hasChangedWebsite ) {
  return !tab.url.includes( NEW_TAB_URL ) && hasChangedWebsite
}

/*
 * shouldShowDashboardOnNewTab – are we on a new tab
 */
async function shouldShowDashboardOnNewTab( tab ) {
  const showDashboardOnNewTab = await getDisplayDashboardOnNewTab()

  return showDashboardOnNewTab && isOnNewTab( tab )
}

/*
 * listeners – so react app and chrome extension can communicate
 */
chrome.runtime.onMessage.addListener(
  function(message, sender, sendResponse) {

    switch( message.type ) {
      case "NAVIGATE_TO_PAGE":
        navigateAllPages();
        break;

      case "FINISH_ONBOARDING":
        finishOnboarding();
        break;

      case "SET_NEXT_PROMPT_TIME":
        const nextPromptTime = parseInt(message.value);
        setNextPromptTime( nextPromptTime );
        break;

      case "SET_DISPLAY_STRATEGY":
        setDisplayStrategy(message.value);
        break;

      case "SET_DISPLAY_DASHBOARD_ON_NEWTAB":
        setDisplayDashboardOnNewTab(message.value);
        break;

      case "SYNC_SUCCESS":
        setIcon('SUCCESS');
        break;

      case "SYNC_FAILURE":
        setIcon('FAILURE');
        break;
    }

  }
);

/*
 * browserAction – opens the dashboard when the icon is clicked
 */
chrome.browserAction.onClicked.addListener( () => {
  chrome.tabs.create({
    url: PAGES.DASHBOARD
  })
})

/*
 * first page to show on installation
 */
function openInstallationPage() {
  chrome.tabs.create({
    url: PAGES.ONBOARDING
  })
}

function currentTime() {
  return Math.round((new Date()).getTime() / 1000);
}

function setNextPromptTime( nextPromptTime ) {
  chrome.storage.sync.set({ nextPromptTime: nextPromptTime }, function() {
    console.log('Time set to '+nextPromptTime);
  });
}

function setDisplayStrategy( displayStrategy ) {
  if( displayStrategy !== 'activeTab' && displayStrategy !== 'newTab' ) {
    displayStrategy = 'activeTab';
  }

  chrome.storage.sync.set({ displayStrategy: displayStrategy }, function() {
    console.log('Display tab set to '+displayStrategy);
  });
}

function setDisplayDashboardOnNewTab( displayDashboardOnNewTab ) {
  chrome.storage.sync.set({ displayDashboardOnNewTab: displayDashboardOnNewTab }, function() {
    console.log('Display dashboard on new tab set to '+displayDashboardOnNewTab);
  });
}

function getDisplayStrategy() {
  return new Promise( (resolve, reject) => {
    chrome.storage.sync.get(['displayStrategy'], ({displayStrategy}) => {
      resolve( displayStrategy );
    });
  })
}

function getDisplayDashboardOnNewTab() {
  return new Promise( (resolve, reject) => {
    chrome.storage.sync.get(['displayDashboardOnNewTab'], ({displayDashboardOnNewTab}) => {
      resolve( displayDashboardOnNewTab );
    });
  })
}

function setIcon(state) {
  let image;

  switch( state ) {
    case "SUCCESS":
      image = "icon_32.png";
      break;
    case "FAILURE":
      image = "notification_icon_32.png"
      break;
    default:
      image = "icon_32.png";
  }

  chrome.browserAction.setIcon({path:`images/${image}`});
}


function setLastHost(tabId, url) {
  let lastHostname = '';

  try {
    url = new URL(url);
    lastHostname = url.hostname;
  } catch (e) {
    lastHostname = (new Date()).getTime().toString();
  }

  chrome.storage.local.get(['lastHostnames'], ({lastHostnames}) => {
    if( lastHostnames === undefined ) {
      lastHostnames = {};
    }

    lastHostnames[tabId] = lastHostname;

    chrome.storage.local.set({ lastHostnames: lastHostnames }, function() {
      console.log('Last hostnames set '+JSON.stringify(lastHostnames));
    });

  });
}

function deleteLastHost(tabId) {
  chrome.storage.local.get(['lastHostnames'], ({lastHostnames}) => {
    if( lastHostnames[tabId] ) {
      delete lastHostnames[tabId];
    }

    chrome.storage.local.set({ lastHostnames: lastHostnames }, function() {
      console.log('Last hostnames set '+JSON.stringify(lastHostnames));
    });

  });
}

function checkPastPromptTime() {
  return new Promise( ( resolve, reject ) => {
    chrome.storage.sync.get(['nextPromptTime'], ({nextPromptTime}) => {
      const hasPast = currentTime() > nextPromptTime;
      resolve( hasPast );
    });
  })
}

function checkWebsiteChange( tabId, newUrl ) {
  return new Promise( ( resolve, reject ) => {
    chrome.storage.local.get(['lastHostnames'], ({lastHostnames}) => {
      newUrl = new URL(newUrl);
      const hasChanged = ( lastHostnames[tabId] !== newUrl.hostname );
      
      resolve( hasChanged );
    });
  })
}

function showApp( tab, url ) {
  chrome.tabs.update(
    tab.id,
    { url: url }
  )
}

/*
 * checkConnection – can the extension connect to Thyself, maybe the user is not on wifi or Thyself is down
 */
function checkConnection() {
  return fetch(`${API_HOST}/ping`)
    .then( response => response.ok )
    .catch( e => false )
}

/*
 * navigateAllPages – move all Thyself screens onto where they wanted to be
 */
function navigateAllPages() {
  const query = { url: `${PAGES.APP}*` };

  chrome.tabs.query( query, ( tabs ) => {
    tabs.forEach( tab => {
      let url = new URL(tab.url);
      const redirectUrl = url.searchParams.get('url') || PAGES.DASHBOARD

      navigate( tab.id, redirectUrl );
    })
  })
}

/*
 * checkWithinTabAllowance – limit the number of Thyself tabs 
 */
function checkWithinTabAllowance() {
  const MAXIMUM_TAB_COUNT = 3;
  const query = { url: [`${PAGES.APP}*`, `${PAGES.DASHBOARD}*`] };

  return new Promise((resolve, reject) => {
    chrome.tabs.query( query, tabs => { 
      const numberTabs = tabs.length
      const withinTabAllowance = numberTabs < MAXIMUM_TAB_COUNT 
      resolve( withinTabAllowance )
    })
  })
}

/*
 * finishOnboarding - move the onboarding page to first mood selection
 */
function finishOnboarding( frame ) {
  const query = { url: `${PAGES.ONBOARDING}*` };

  chrome.tabs.query( query, ( tabs ) => {
    tabs.forEach( tab => {
      navigate( tab.id, PAGES.APP );
    })
  })
}

function navigate( tabId, url ) {
  chrome.tabs.update( tabId, { url: url } )
}
